
# NAF Swarm

- Goal is to convert an AST composed of AND and XOR nodes to NAF.
- NAF is one of:
  - VHL: V AND HI XOR LO
  - AND
  - SUM: (generalized XOR)

In general, we set it up so that for VHL nodes, the variable is the topmost
variable in the whole sub-expression.

- the solver strategy
  - refine one ANF term at a time.
  - loop:
    - walk downward to find the bottom-most refined VHL of the term
    - the children are necessarily AND or SUM nodes
    - pick one child to refine
    - using a depth first walk, rewrite every subnode into VHL form.
    - the next highest variable should rise up to form a new base for the term.

## Refining a SUM node

We need a way to mark that a term is already exposing its highest variable.

This is a simple as finding the term with the highest top variable.
Move the other terms into a new SUM node.
AND the top variable with the rest of the term it came from.

## Refining an AND node

This was already implemented for the initial AST -> NAF conversion.
As long as we bring the inputs into ANF form, it should be fine.


## Using VhlSwarm

The WIP system can already handle all this for us.

The WIP system essentially gives us a big scaffold to hold everything in place
while the swarm refines each subcomponent.

## How can I visualize this?
- Render the initial AST for a small function using GSL
- Option 1: Try to visualize current NafBase, slowly migrate to VhlSwarm.
- Option 2: Visualize empty VhlSwarm from the start, and rewrite Naf in this form
  - Define the VhlSwarm types:
    - J: would be enum { VHL|AND|SUM }, right?
    - Where do the nids point, though? I think we still need .nodes
    - Will we add to .nodes? No? I think everything we add is just WIP.
  - Right now, by the time we go to create the top node, we've already built everything from the bottom up
  - For the second one, we don't have that luxury because we're no longer using an AST.
  - What if we just kept it as an AST?
  - Well, we just have to walk downward, marking dependencies, and then walk upwards, converting to VHL.
  - We can't convert to VHL from the top down because we don't know what V is at the top.
  - Also, if we're going to distribute this, we have to do it with a data structure that supports the swarm.
  - We know the WorkState works fine with a swarm, so we should just start there.
  - Assume we have a read-only copy of the original AST.
  - Now, we want to convert a particular node to Vhl form. (So "ToVHL(NID)" is one of our job types)
  - That recursively converts a bunch of other nodes to Vhl form.
  - Once the child nodes are converted to VHL (and this is cached), we do the AND or XOR.
  - On each bottom-up pass, VHLs stay in place, and other nodes get turned into VHLs
    - which also creates a bunch of extra nodes
  - what if every step looked like the first step?
    - given a node, first mark its dependencies from the top down
    - next, rebuild the AST from the bottom up
    - on the way up, push each item into a queue so the swarm handles it
      - vhl nodes don't need any work, they just get copied over directly
      - for AND and SUM, we have to decide what we know about them:
        - they may be a dependency we have to translate (send to swarm)
        - or they may be referred to by a VHL that doesn't require work yet
          (so we have to keep them but not modify)
        - or they may not be marked as relevant at all (garbage collected/ignored)
      - so swarm takes AND and SUM nodes that actually need work
        swarm returns a vec of new AST nodes for each old node

We could do some work up-front to color AST nodes based on their highest variables.
 - Could just use a bitmask since we're only concerned with < 64 variables.
 - But what does this buy us? I guess it would pre-search the nodes we don't care about.
 - But VHL nodes already serve as a good boundary for this.

I guess I'm thinking about how to avoid having to do the same mark and sweep every single time.
 - What if the swarm just recursively handled this?


We could isolate any expression that concerns only the "lower" variables, and move these to their own AST.
- This really is just to simplify our work, since we will never touch these AST nodes.
- Next, we continuously rewrite the AST to put it into ANF form.
- We always try to rewrite nodes in place to move them into ANF/VHL form.


We either have to track dependencies on the way down, or we have to linearize it on the way up.
(By linearize i mean arrange to push the bottom-up build into a queue on the way up.)

With the top-down approach, we have the potential to someday implement short circuiting: x and 0 is 0, so if a 0 bubbles up to an AND, we can cancel the other side. AFAIK, this is not yet implemented in the WIP base.

Jobs ought to be like:
  - TopVar(x) [But we can't cache this, because it changes]
  - ToTop(v,nid) -> this we can cache.


I think we should just make a function like this:
  AST -> AST
  But have it use the swarm to build it at each step.
  The swarm itself can build a new AST in fresh AstBase.
  It would be nice if an AST could refer to nodes in another AST.
  (Namespaces in nids would allow this.)



## visualization plan
- Make a tool to visualize (small) ASTs
- Start with a plain AST, sorted and packed appropriately.
- For each round:
  - Figure out which nodes might change.
  - Mark dependencies backwards to know which ones we still need.
  - Build new AST from bottom up, with main thread delegating each node to swarm.
  - Main thread sends swarm a message, gets back a new AST to append.
- Visualization:
  - LHS: area to visualize old AST nodes
  - col0: a vertical display of the AST as Op(*args)
  - col1: the corresponding new nodes (multiple rows per col0)
  - RHS: area to visualize new AST nodes
  - stepper to build the new AST



## Why is this so different from BddSwarm?
- With BddSwarm, we already have everything in ITE form, and we're just combining them in different ways.
- (Mostly because we're starting by translating AST nodes into trivial BDD variables and then replacing them)




## What changes to make this distributed?

- If we use this as a general-purpose ANF solver, it's just like BddSwarm but using (lo~:v*.hi)
- But NAF is weird because we can also use it in a partial form.
  - In this form, we can grow terms we're interested in downward from the top.
  - Everything will be WIP all the time (we basically never complete a node)
  - Top nodes are VHL but the H and L are unresolved.
  - Once a node is in in VHL form, it's "done" except to wait for child nodes to fully resolve.
  - All other nodes are in "AND" or "SUM" form.

- We seem to be missing something that turns AND->VHL and SUM->VHL
  well, i guess that's nafbase.vhl(), but i don't see how it's called.
